/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// Package pem encapsulates the management of TLS Secret contents as
// PEM files -- files in which the certificate and key are
// concatenated.
package pem

import (
	"crypto/sha512"
	"encoding/base64"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"

	"golang.org/x/sys/unix"
)

// File encapsulates the metadata of a PEM file -- the k8s Secret that
// it represents, and information about the file.
type File struct {
	Namespace       string
	Name            string
	UID             string
	ResourceVersion string
	Size            int64
	ModTime         time.Time
	Hash            [sha512.Size]byte
}

func (pem File) String() string {
	return pem.Namespace + "_" + pem.Name
}

func (pem File) hashedName() string {
	hash := sha512.New512_256()
	hash.Write([]byte(pem.Namespace))
	hash.Write([]byte(pem.Name))
	bytes := hash.Sum(nil)
	return base64.RawURLEncoding.EncodeToString(bytes)
}

// Basename returns a file basename for the file. The name is
// "<Namespace>_<Name>.pem", unless this results in a name that is
// longer than NAME_MAX on the current system. In the latter case, the
// name is an opaque hash with the ".pem" suffix, guaranteed to be
// shorter than NAME_MAX.
func (pem File) Basename() string {
	name := pem.String()
	if len(name)+len(".pem") > unix.NAME_MAX {
		name = pem.hashedName()
	}
	return name + ".pem"
}

// Files encapsulates metadata about all of the PEM files that are
// currently stored on the system, and provides capabilites to manage
// the files.
//
// Base is the name of the base directory in which PEM files are
// stored.
//
// If Gid != -1, then it is a group ID set in the permissions of PEM
// files.
//
// Operations that read contents of the metadata and file system are
// protected by a shared mutex (read lock). Operations that change
// contents of the metadata or file system are protected by an
// exclusive lock (write lock).
type Files struct {
	Files     map[string]*File
	Base      string
	Gid       int
	mtx       *sync.RWMutex
	crtGetter *crt.Getter
}

// NewFiles returns a Files instance. base MUST be set to the name of
// a directory that is writable by the current process owner.  If gid
// is >= 0, then it is the group id set in PEM file permissions.
// crtGetter is used to retrieve and validate the contents of TLS
// Secrets.
func NewFiles(base string, gid int, crtGetter *crt.Getter) (*Files, error) {
	if crtGetter == nil {
		return nil, fmt.Errorf("crtGetter is nil")
	}
	info, err := os.Stat(base)
	if err != nil {
		return nil, fmt.Errorf("Cannot stat %s: %v", base, err)
	}
	if !info.IsDir() {
		return nil, fmt.Errorf("%s: not a directory", base)
	}
	if err = unix.Access(base, unix.W_OK); err != nil {
		return nil, fmt.Errorf("%s not writable: %v", base, err)
	}
	files := &Files{
		Base:      base,
		Gid:       -1,
		Files:     make(map[string]*File),
		mtx:       new(sync.RWMutex),
		crtGetter: crtGetter,
	}
	if gid >= 0 {
		files.Gid = gid
	}
	return files, nil
}

func (pemfiles *Files) String() string {
	s := make([]string, len(pemfiles.Files)+2)
	s[0] = "base:" + pemfiles.Base
	s[1] = "gid:" + strconv.Itoa(pemfiles.Gid)
	i := 2
	for k := range pemfiles.Files {
		s[i] = k
	}
	return "{" + strings.Join(s, ",") + "}"
}

// Path returns the file path to be used for the Secret represented by
// pem.
func (pemfiles *Files) Path(pem *File) string {
	return filepath.Join(pemfiles.Base, pem.Basename())
}

// Have returns true iff the Secret designated by namespace and name,
// and optionally uid and/or version, is present in the current
// metadata.
//
// namespace and name MUST be non-empty. Either or both of uid or
// version MAY be empty. If uid and/or version are non-empty, then
// true is returned only if they match the coreesponding UID and/or
// ResourceVersion fields.
//
// Have is protected by a read lock.
func (pemfiles *Files) Have(namespace, name, uid, version string) bool {
	pemfiles.mtx.RLock()
	defer pemfiles.mtx.RUnlock()

	key := namespace + "/" + name
	pem, exists := pemfiles.Files[key]
	if !exists {
		return false
	}

	if namespace != pem.Namespace || name != pem.Name {
		panic("name/namespace mismatch: " + pemfiles.String())
	}

	if uid != "" && uid != pem.UID {
		return false
	}
	if version != "" && version != pem.ResourceVersion {
		return false
	}
	return true
}

func mkPEM(crt, key []byte) []byte {
	pem := make([]byte, 0, len(crt)+len(key)+1)
	pem = append(pem, crt...)
	if crt[len(crt)-1] != byte('\n') {
		crt = append(crt, byte('\n'))
	}
	return append(crt, key...)
}

// Check returns true iff the current metadata for a Secret match
// the current contents of the PEM file.
//
// ns and name MUST be non-empty. Either or both of uid or version MAY
// be empty. If uid and/or version are non-empty, then true is
// returned only if they match the coreesponding UID and/or
// ResourceVersion fields.
//
// If a matching entry in the metadata is found, then true is returned
// iff the current values of certificate and key in the Secret match
// the current contents of the file, and the file metadata match the
// file currently stored on the system -- size, modification time, and
// hash of the contents.
//
// Moreover, the file must be group-readable, and if pemfiles.Gid !=
// -1, then it must match the group ID in the files permissions.
//
// Check is protected by a read lock.
func (pemfiles *Files) Check(ns, name, uid, version string) bool {
	pemfiles.mtx.RLock()
	defer pemfiles.mtx.RUnlock()

	mapKey := ns + "/" + name
	pem, exists := pemfiles.Files[mapKey]
	if !exists {
		return false
	}
	crt, key, _, _, found, valid, err := pemfiles.crtGetter.Get(
		ns, name, uid, version)
	if !found || !valid {
		return false
	}
	if err != nil {
		return false
	}
	if sha512.Sum512(mkPEM(crt, key)) != pem.Hash {
		return false
	}
	path := pemfiles.Path(pem)
	info, err := os.Stat(path)
	if err != nil {
		return false
	}
	if info.Size() != pem.Size {
		return false
	}
	if info.ModTime() != pem.ModTime {
		return false
	}
	if info.Mode().Perm()&040 == 0 {
		return false
	}
	if pemfiles.Gid >= 0 {
		ustat := &unix.Stat_t{}
		if err = unix.Stat(path, ustat); err != nil {
			return false
		} else if int(ustat.Gid) != pemfiles.Gid {
			return false
		}
	}
	if contents, err := ioutil.ReadFile(path); err != nil {
		return false
	} else if sha512.Sum512(contents) != pem.Hash {
		return false
	}
	return true
}

// Write saves the contents of a TLS Secret in a PEM file, if the
// Secret is found and its contents are valid.
//
// ns and name MUST be non-empty. Either or both of uid or version MAY
// be empty. If uid and/or version are non-empty, then the Secret is
// found only if they match the Secrets's coreesponding UID and/or
// ResourceVersion fields.
//
// The return value found is true iff the Secret is stored by the
// k8s lister used to initialize the crt.Getter for the Files object.
//
// If the Secret is found, then it is validated (see the Getter function
// in package pkg/crt). The return value valid is true iff the Secret
// is a valid TLS Secret.
//
// No more than one of the return values found and valid is false. The
// write operation is successful if both found and valid are true, and
// err is non-nil. The resulting file is group-readable but not
// accessible to other (file mode 0640). If pemfiles.Gid != -1, then
// the group ID in the file's permissions is set accordingly.
//
// This operation may create a new file, or overwrite an existing file
// -- file creation and update are not distinguished.
//
// Write is protected by a write lock.
func (pemfiles *Files) Write(
	ns, name, uid, version string,
) (found, valid bool, err error) {
	crt, key, suid, sversion, found, valid, err := pemfiles.crtGetter.Get(
		ns, name, uid, version)
	if !found || !valid || err != nil {
		return
	}

	pemfile := &File{
		Namespace:       ns,
		Name:            name,
		UID:             suid,
		ResourceVersion: sversion,
	}
	pem := mkPEM(crt, key)
	path := pemfiles.Path(pemfile)

	pemfiles.mtx.Lock()
	defer pemfiles.mtx.Unlock()

	if err = ioutil.WriteFile(path, pem, os.FileMode(0640)); err != nil {
		return
	}
	if pemfiles.Gid >= 0 {
		if err = os.Chown(path, -1, pemfiles.Gid); err != nil {
			return
		}
	}
	info, er := os.Stat(path)
	if er != nil {
		err = er
		return
	}
	pemfile.Size = info.Size()
	pemfile.ModTime = info.ModTime()
	pemfile.Hash = sha512.Sum512(pem)
	pemfiles.Files[ns+"/"+name] = pemfile
	return found, valid, nil
}

// Delete deletes the PEM file coreesponding to the TLS Secret
// designated by namespace and name, if the file exists.
//
// The boolean return value is true iff metadata was stored for the
// PEM file, and the file was stored on the system before the
// attempted delete.
//
// The error return value is non-nil for any error deleting the file.
// If os.IsNotExist() is true of a non-nil error, then the boolean
// return value is false.
//
// Delete is protected by a write lock.
func (pemfiles *Files) Delete(namespace, name string) (bool, error) {
	pemfiles.mtx.Lock()
	defer pemfiles.mtx.Unlock()

	key := namespace + "/" + name
	pem, exists := pemfiles.Files[key]
	if !exists {
		return false, nil
	}
	path := pemfiles.Path(pem)

	if err := os.Remove(path); err != nil {
		return !os.IsNotExist(err), err
	}
	delete(pemfiles.Files, key)
	return true, nil
}

// GetFileInfo returns information suitable for JSON marshaling about
// a PEM file, if a file is currently stored for the given namespace
// and name, and optionally matching the UID and ResourceVersion.
//
// ns and name MUST be set, while uid and/or version MAY be empty. If
// either of uid or version is set, then the FileInfo is returned only
// if the UID and/or ResourceVersion of the corresponding Secret
// match.
//
// The found return value is true iff the corresponding PEM file is
// currently stored.
//
// If an entry is found, the PEM is read and its contents are parsed
// to populate the Cert field, to extent supported by go's crypto/x509
// facilities.
//
// GetFileInfo is protected by a read lock.
func (pemfiles *Files) GetFileInfo(
	ns, name, uid, version string,
) (info FileInfo, found bool, err error) {
	mapKey := ns + "/" + name

	pemfiles.mtx.RLock()
	defer pemfiles.mtx.RUnlock()

	pemFile, exists := pemfiles.Files[mapKey]
	if !exists {
		err = fmt.Errorf("%s not found", mapKey)
		return
	}
	if uid != "" && pemFile.UID != uid {
		err = fmt.Errorf("%s uid=%s not found", mapKey, uid)
		return
	}
	if version != "" && pemFile.ResourceVersion != version {
		err = fmt.Errorf("%s version=%s not found", mapKey, version)
		return
	}

	path := pemfiles.Path(pemFile)
	pemBytes, err := ioutil.ReadFile(path)
	if err != nil {
		if !os.IsNotExist(err) {
			found = true
		}
		err = fmt.Errorf("%s: cannot read pem file: %v", mapKey, err)
		return
	}
	found = true

	crt, err := mkCrt(pemBytes)
	if err != nil {
		err = fmt.Errorf("%s: cannot read certificate: %v", mapKey, err)
		return
	}
	info.K8sSecret = Secret{
		Namespace: pemFile.Namespace,
		Name:      pemFile.Name,
		UID:       pemFile.UID,
		Version:   pemFile.ResourceVersion,
	}
	info.Meta = FileMeta{
		ModTime: pemFile.ModTime,
		Size:    pemFile.Size,
	}
	info.Cert = crt
	return
}

// GetAllFileInfo returns information about all of the PEM files
// currently stored, suitable for JSON marshaling. Is is protected by
// a read lock.
func (pemfiles *Files) GetAllFileInfo() ([]FileInfo, error) {
	allInfo := make([]FileInfo, 0, len(pemfiles.Files))

	pemfiles.mtx.RLock()
	defer pemfiles.mtx.RUnlock()

	for _, pemFile := range pemfiles.Files {
		info, _, err := pemfiles.GetFileInfo(
			pemFile.Namespace, pemFile.Name, "", "")
		if err != nil {
			return allInfo, err
		}
		allInfo = append(allInfo, info)
	}
	return allInfo, nil
}
